/*
 * Copyright (c) 2005-2010 KOM � Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.component.relevance.vehicular.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.component.relevance.vehicular.ImpactFunction;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.vectoral.VectoralProperty;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.information.AggregatedInformation;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.information.AvailableInformationAttributes;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.information.PointInformation;

public abstract class AbstractQoIBasedImpactFunction<T extends PointInformation> implements ImpactFunction<T> {
	protected static final long SCALING = Time.SECOND;
	protected static final double ACCURACY_FACTOR = 100000;

	protected double accuracy;
	protected double costWrongKeep = 1;
	protected double costWrongChange = 1;
	protected double b;
	protected double rate;
	protected double numberOfMessages;
	protected long ttl;
	protected long maxTimestamp;
	protected long minTimestamp;
	protected int maxCacheSize;

	protected Object _lastDecision;


	public AbstractQoIBasedImpactFunction(List<T> pInformation, double pAccuracy, int pMaxCacheSize) {
		this(pInformation, pAccuracy, pMaxCacheSize, null);
	}

	public AbstractQoIBasedImpactFunction(List<T> pInformation, double pAccuracy, int pMaxCacheSize,
			Object pLastDecision) {
		accuracy = pAccuracy;
		maxCacheSize = pMaxCacheSize;
		_lastDecision = pLastDecision;
		minTimestamp = Long.MAX_VALUE;
		maxTimestamp = 0;
		Object value = pInformation.get(0).getValue();
		if (value instanceof VectoralProperty) {
			value = ((VectoralProperty) value).getMostProbableIndex();
		}

		List<Object> possibleValues = new ArrayList<>();

		for (T t : pInformation) {
			if (!t.hasAttribute(AvailableInformationAttributes.TTL)) {
				throw new AssertionError("Unable to perform TTL-based majority voting witout TTL");
			}
			long timestamp = t.getDetectionDate();

			if (timestamp < minTimestamp) {
				minTimestamp = timestamp;
			}
			if (timestamp > maxTimestamp) {
				maxTimestamp = timestamp;
			}
			if (pInformation instanceof AggregatedInformation) {
				if (((AggregatedInformation) pInformation).isAggregated()) {
					long temp = ((AggregatedInformation) pInformation).getAggregationInformation().getMinTimestamp();
					if (temp < minTimestamp) {
						minTimestamp = temp;
					}
				}
			}

			Object currentValue = t.getValue();
			if (currentValue instanceof VectoralProperty) {
				VectoralProperty currentProperty = (VectoralProperty) currentValue;
				currentValue = currentProperty.getMostProbableIndex();

				for (int i = 0; i < currentProperty.getValueProbabilities().length; i++) {
					if (!possibleValues.contains(i)) {
						possibleValues.add(i);
					}
				}
			} else {
				if (!possibleValues.contains(currentValue)) {
					possibleValues.add(currentValue);
				}
			}
		}

		long difference = maxTimestamp - minTimestamp;

		int amount = 0;
		for (T t : pInformation) {
			if (t instanceof AggregatedInformation && ((AggregatedInformation) t).isAggregated()) {
				amount += ((AggregatedInformation) t).getAggregationInformation().getNumberOfIndiviualMeasurements();
			} else {
				amount++;
			}
		}

		ttl = getTTL(pInformation.get(0));
		numberOfMessages = amount * ttl * SCALING / (difference + Time.SECOND);

		rate = ttl / (double) numberOfMessages;

		List<Double> bValues = new ArrayList<>();
		for (Object possibleValue : possibleValues) {
			double temp = determineB(pInformation.get(0).getValue(), possibleValue,
					getChangeRate(pInformation.get(0), rate), rate, 1 - pAccuracy,
					numberOfMessages, costWrongKeep, costWrongChange);

			if (!Double.isNaN(temp)) {
				bValues.add(temp);
			}
		}

		Collections.sort(bValues);

		if (bValues.size() > 0) {
			if (bValues.size() % 2 == 0) {
				b = (bValues.get(bValues.size() / 2) + bValues.get(bValues.size() / 2 - 1)) / 2.0;
			} else {
				b = bValues.get(bValues.size() / 2);
			}
		} else {
			b = Double.NEGATIVE_INFINITY;
		}
	}

	protected abstract double determineB(Object pExampleValue, Object pPossibleValue, double pChangeRate, double pRate,
			double pErrorProbability, double pNumberOfMessages, double pCostWrongKeep, double pCostWrongChange);

	@Override
	public double calculateImpact(long pDetectionDate) {
		double impact = calculateImpact(1 - accuracy, numberOfMessages,
				((ttl - (maxTimestamp - pDetectionDate) / SCALING) / (double) ttl) * numberOfMessages, b) / (accuracy);
		return impact;
	}

	protected <T extends PointInformation> long getTTL(T pT) {
		return (long) pT.getAttribute(AvailableInformationAttributes.TTL) / SCALING;
	}

	protected <T extends PointInformation> double getChangeRate(T pT, double pRate) {
		return getChangeProbability((long) (getTTL(pT) / pRate));
	}

	protected double getChangeProbability(long ttl) {
		return 1 - Math.pow(0.5, 1 / (double) ttl);
	}

	protected int getOptimalMessageAmountForSwitch(double changeProbability, double errorProbability, double costSlow,
			double costFast) {
		return (int) Math.round(Math.log(-changeProbability / Math.log(errorProbability) * costSlow / costFast)
				/ Math.log(errorProbability));
	}

	public double calculateImpact(double errorProbability, double pNumberOfMessages, double pMessageNumber, double b) {
		double age = pNumberOfMessages - pMessageNumber;
		if (errorProbability == 0) {
			if (pMessageNumber == pNumberOfMessages) {
				return 1;
			} else {
				return 0;
			}
		} else if (errorProbability == 1) {
			return 1;
		} else if (errorProbability == 0.5 || b == 0) {
			return (1 - errorProbability) / pNumberOfMessages * age + errorProbability;
		} else if (b == Double.NEGATIVE_INFINITY) {
			if (pMessageNumber == pNumberOfMessages) {
				return 1 - errorProbability;
			} else {
				return 0;
			}
		} else if (b == Double.POSITIVE_INFINITY) {
			return 1;
		}

		return (1 - errorProbability) * (Math.exp(b * age) - Math.exp(b * pNumberOfMessages))
				/ (1 - Math.exp(b * pNumberOfMessages));
	}

}
