/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


package de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.vectoral;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.ProbabilityDistribution;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.TemporalDependencyMatrix;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.AbstractRoadProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.Location;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkEdge;

/**
 * @author Tobias Meuser (tobias.meuser@kom.tu-darmstadt.de)
 * @version 1.0 at 27.02.2018
 *
 */
public abstract class VectoralProperty extends AbstractRoadProperty implements Cloneable {
	private static Map<DeviationIdentifier, Double> _requiredStandardDeviation = new HashMap<>();

	private double[] _valueProbabilities;

	public VectoralProperty(Location pLocation, RoadNetworkEdge pEdge, int valueAmount) {
		this(pLocation, pEdge, createProbabilityArray(valueAmount));
	}

	public VectoralProperty(Location pLocation, RoadNetworkEdge pEdge, double[] values) {
		super(pLocation, pEdge);

		_valueProbabilities = values;
	}

	public void setProbabilities(double[] pValueProbabilities) {
		_valueProbabilities = pValueProbabilities;
	}

	protected void setGaussianProbabilities(double pMean, double pDeviation) {
		if (pDeviation > 0) {
			double sum = 0;
			for (int i = 0; i < _valueProbabilities.length; i++) {
				double probability = 1 / (2 * Math.PI * Math.pow(pDeviation, 2))
						* Math.exp(-Math.pow(i - pMean, 2) / (2 * Math.pow(pDeviation, 2)));
				_valueProbabilities[i] = probability;
				sum += probability;
			}

			for (int i = 0; i < _valueProbabilities.length; i++) {
				_valueProbabilities[i] /= sum;
			}
		} else {
			Arrays.fill(_valueProbabilities, 0);

			_valueProbabilities[(int) Math.round(pMean)] = 1;
		}
	}

	public static double[] createProbabilityArray(int valueAmount) {
		double[] probabilities = new double[valueAmount];
		Arrays.fill(probabilities, 1 / (double) probabilities.length);
		return probabilities;
	}

	public double[] getValueProbabilities() {
		return _valueProbabilities;
	}

	public abstract Object getValueAtIndex(int pIndex);

	public int getMostProbableIndex() {
		double max = -1;
		int index = 0;

		double[] probs = getValueProbabilities();
		for (int i = 0; i < probs.length; i++) {
			if (probs[i] > max) {
				max = probs[i];
				index = i;
			}
		}

		return index;
	}

	public Object getMostProbableValue() {
		return getValueAtIndex(getMostProbableIndex());
	}

	public abstract int getIndexForValue(Object pValue);

	public abstract TemporalDependencyMatrix getDependencyMatrix();

	@Override
	public abstract VectoralProperty clone();

	public VectoralProperty age(long pAge) {
		return age(pAge, getDependencyMatrix());
	}

	public VectoralProperty age(long pAge, TemporalDependencyMatrix pTemporalDependency) {
		VectoralProperty result = clone();

		double[] valueProbabilities = result.getValueProbabilities();
		double[] currentProbabilities = getValueProbabilities();

		for (int t = 0; t < pAge; t++) {

			Arrays.fill(valueProbabilities, 0);

			for (int i = 0; i < valueProbabilities.length; i++) {
				for (int j = 0; j < valueProbabilities.length; j++) {
					double added = currentProbabilities[i] * pTemporalDependency.getTransitionProbability(i, j);
					valueProbabilities[j] += added;
				}
			}

			double sum = 0;
			for (int i = 0; i < valueProbabilities.length; i++) {
				sum += valueProbabilities[i];
			}
			for (int i = 0; i < valueProbabilities.length; i++) {
				valueProbabilities[i] /= sum;
			}

			currentProbabilities = Arrays.copyOf(valueProbabilities, valueProbabilities.length);
		}

		return result;
	}

	public double getProbabilityForIndex(int pIndex) {
		return getValueProbabilities()[pIndex];
	}

	public abstract double getProbabilityForValue(Object pValue);

	public VectoralProperty combine(VectoralProperty pVectoralProperty) {
		VectoralProperty result = clone();

		double sum = 0;

		double[] valueProbabilities = result.getValueProbabilities();
		for (int i = 0; i < valueProbabilities.length; i++) {
			valueProbabilities[i] *= pVectoralProperty.getValueProbabilities()[i];
			sum += valueProbabilities[i];
		}

		for (int i = 0; i < valueProbabilities.length; i++) {
			valueProbabilities[i] /= sum;
		}

		return result;
	}

	public void set(int pIndex, double pAccuracy, ProbabilityDistribution pDist) {
		switch (pDist) {
		case SIMPLE:
			setSimple(pIndex, pAccuracy);
			break;
		case GAUSSIAN:
			setGaussianWithAccuracy(pIndex, pAccuracy);
			break;
		default:
			throw new AssertionError("Unknown ProbablilityDistribution " + pDist);
		}
	}

	public void setSimple(int pIndex, double pAccuracy) {
		double[] valueProbabilities = getValueProbabilities();
		Arrays.fill(valueProbabilities, 0);
		int index = (int) (pIndex);
		valueProbabilities[index] += pAccuracy;

		int distance = 1;

		if (index + distance < valueProbabilities.length) {
			valueProbabilities[index + distance] += (1 - pAccuracy) / 2.0;
		} else {
			valueProbabilities[index - distance] += (1 - pAccuracy) / 2.0;
		}

		if (index - 1 >= 1) {
			valueProbabilities[index - distance] += (1 - pAccuracy) / 2.0;
		} else {
			valueProbabilities[index + distance] += (1 - pAccuracy) / 2.0;
		}
	}

	public void setGaussianWithAccuracy(double pValue, double pAccuracy) {
		double deviation = determineCorrectDeviation((int) pValue, pAccuracy);
		setGaussianProbabilities(pValue, deviation);
	}

	public void setGaussian(double pIndex, double pDeviation) {
		setGaussianProbabilities(pIndex, pDeviation);
	}

	public double determineCorrectDeviation(int pIndex, double pAccuracy) {
		DeviationIdentifier identifier = new DeviationIdentifier(getClass(), pIndex, pAccuracy);

		if (!_requiredStandardDeviation.containsKey(identifier)) {
			if (pAccuracy < 1) {
				VectoralProperty vectoralProperty = clone();

				double deviation = 10;

				boolean reduceChange = false;
				double currentChange = deviation / 2;

				do {
					vectoralProperty.setGaussian(pIndex, deviation);
					if (vectoralProperty.getProbabilityForIndex((int) pIndex) > pAccuracy) {
						deviation += currentChange;
					} else {
						deviation -= currentChange;
						reduceChange = true;
					}

					if (reduceChange) {
						currentChange /= 2.0;
					}
				} while (Math.round(vectoralProperty.getProbabilityForIndex(pIndex) * 10000) != Math
						.round(pAccuracy * 10000));

				_requiredStandardDeviation.put(identifier, deviation);
			} else {
				_requiredStandardDeviation.put(identifier, 0d);
			}
		}

		return _requiredStandardDeviation.get(identifier);
	}

	public abstract int getDefaultIndex();

	@Override
	public boolean equals(Object pObj) {
		if (pObj instanceof VectoralProperty) {
			VectoralProperty vector = (VectoralProperty) pObj;
			
			return vector.getClass().equals(getClass()) && vector.getEdge().equals(getEdge())
					&& vector.getMostProbableIndex() == getMostProbableIndex();
		}
		return super.equals(pObj);
	}

	private class DeviationIdentifier {
		private static final double GRANULARITY = 0.0001;

		private Class<? extends VectoralProperty> _propertyClass;
		private int _index;
		private double _accuracy;

		public DeviationIdentifier(Class<? extends VectoralProperty> pPropertyClass, int pIndex, double pAccuracy) {
			_propertyClass = pPropertyClass;
			_index = pIndex;
			_accuracy = (int) (pAccuracy / GRANULARITY) * GRANULARITY;
		}

		public Class<? extends VectoralProperty> getPropertyClass() {
			return _propertyClass;
		}

		public int getIndex() {
			return _index;
		}

		public double getAccuracy() {
			return _accuracy;
		}

		@Override
		public int hashCode() {
			int prime = 31;
			return Integer.hashCode(_index) * prime + Double.hashCode(_accuracy);
		}

		@Override
		public boolean equals(Object obj) {
			if (obj instanceof DeviationIdentifier) {
				DeviationIdentifier deviationIdentifier = (DeviationIdentifier) obj;
				return deviationIdentifier._accuracy == _accuracy && deviationIdentifier._index == _index
						&& deviationIdentifier._propertyClass.equals(_propertyClass);
			}
			return false;
		}

		private VectoralProperty getOuterType() {
			return VectoralProperty.this;
		}
	}

	@Override
	public Object getValue() {
		return getMostProbableValue();
	}

	@Override
	public String toString() {
		return Arrays.toString(_valueProbabilities);
	}
}
