/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


package de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.roadcondition;

import java.util.Arrays;

import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.ProbabilityDistribution;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.TemporalDependencyMatrix;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.jam.EventInformationContainer;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.RoadProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.vectoral.VectoralProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.Location;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkEdge;

/**
 * @author Tobias Meuser (tobias.meuser@kom.tu-darmstadt.de)
 * @version 1.0 at 12.04.2018
 *
 */
public class VectoralRoadConditionProperty extends VectoralProperty {
	private static final int DIMENSIONS = 6;

	private static TemporalDependencyMatrix _temporalDependencyMatrix;

	public VectoralRoadConditionProperty(Location pLocation, RoadNetworkEdge pEdge) {
		super(pLocation, pEdge, DIMENSIONS);
	}

	/**
	 * 
	 */
	public VectoralRoadConditionProperty() {
		this(null, null);
	}

	public void setConditon(RoadCondition pCondition, double pAccuracy, ProbabilityDistribution pDist) {
		switch (pDist) {
		case SIMPLE:
			setSimple(pCondition.ordinal(), pAccuracy);
			break;
		case GAUSSIAN:
			setGaussianWithAccuracy(pCondition.ordinal(), pAccuracy);
			break;
		default:
			throw new AssertionError("Unknown ProbablilityDistribution " + pDist);
		}
	}

	@Override
	public VectoralRoadConditionProperty clone() {
		VectoralRoadConditionProperty vectoralJamProperty = new VectoralRoadConditionProperty(getLocation(), getEdge());
		vectoralJamProperty.setProbabilities(Arrays.copyOf(getValueProbabilities(), getValueProbabilities().length));
		vectoralJamProperty.setDetectionDate(getDetectionDate());
		vectoralJamProperty.setDuration(getDuration());
		return vectoralJamProperty;
	}

	@Override
	public VectoralRoadConditionProperty age(long pAge) {
		return (VectoralRoadConditionProperty) super.age(pAge);
	}

	@Override
	public RoadProperty getDefaultProperty() {
		return new VectoralRoadConditionProperty(getLocation(), getEdge());
	}

	@Override
	public RoadCondition getValueAtIndex(int pIndex) {
		return RoadCondition.values()[pIndex];
	}

	@Override
	public int getIndexForValue(Object pValue) {
		return RoadCondition.valueOf(pValue.toString()).ordinal();
	}

	@Override
	public TemporalDependencyMatrix getDependencyMatrix() {
		if (_temporalDependencyMatrix == null) {
			TemporalDependencyMatrix temporalDependencyMatrix = new TemporalDependencyMatrix(DIMENSIONS);
			double TEMPORAL_CHANGE = 1
					/ ((double) EventInformationContainer.getEventDuration(RoadConditionProperty.class) / Time.SECOND)
					/ 2.0;

			{
				double[] probabilities = new double[DIMENSIONS];
				probabilities[0] += (1 - (DIMENSIONS - 1) * TEMPORAL_CHANGE);
				for (int x = 1; x < DIMENSIONS - 1; x++) {
					probabilities[x] += TEMPORAL_CHANGE;
				}
				temporalDependencyMatrix.setTransitionProbability(0, probabilities);
			}

			for (int x = 1; x < DIMENSIONS; x++) {
				double[] probabilities = new double[DIMENSIONS];
				if (x > 1) {
					probabilities[x - 1] += TEMPORAL_CHANGE;
				} else {
					probabilities[x + 1] += TEMPORAL_CHANGE;
				}
				if (x < DIMENSIONS - 1) {
					probabilities[x + 1] += TEMPORAL_CHANGE;
				} else {
					probabilities[x - 1] += TEMPORAL_CHANGE;
				}
				probabilities[x] += (1 - 2 * TEMPORAL_CHANGE);

				temporalDependencyMatrix.setTransitionProbability(x, probabilities);
			}

			_temporalDependencyMatrix = temporalDependencyMatrix;
		}
		return _temporalDependencyMatrix;
	}

	public double getProbabilityForValue(Object pValue) {
		if (pValue instanceof RoadCondition) {
			RoadCondition value = (RoadCondition) pValue;

			int index = value.ordinal();
			return getValueProbabilities()[index];
		}
		return 0;
	}
	
	@Override
	public int getDefaultIndex() {
		return 0;
	}
}