/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


package de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.plugin;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import de.tudarmstadt.maki.simonstrator.api.Host;
import de.tudarmstadt.maki.simonstrator.api.Randoms;
import de.tudarmstadt.maki.simonstrator.api.component.ComponentNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.ProbabilityDistribution;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.measurement.MeasurementDistributionTypeContainer;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.EnvironmentProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.JamProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.RoadProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.vectoral.VectoralJamProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.Location;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSComponent;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSRequest;
import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.InformationNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.sis.type.SiSTypes;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkEdge;
import de.tudarmstadt.maki.simonstrator.api.util.XMLConfigurableConstructor;

/**
 * @author Tobias Meuser (tobias.meuser@kom.tu-darmstadt.de)
 * @version 1.0 at 27.02.2018
 *
 */
public class MaliciousVectoralJamEnvironmentSensorPlugin implements EnvironmentSensorPlugin {
	private Host _host;

	private SiSComponent _sis;

	private double _accuracy;

	private final double _originalAccuracy;
	private final double _maxDeviation;

	private double _valueOffset;
	private double _communicatedAccuracy;

	private static Random _random = Randoms.getRandom(MaliciousVectoralJamEnvironmentSensorPlugin.class);

	private static Random _accuracyRandom = Randoms.getRandom(MaliciousVectoralJamEnvironmentSensorPlugin.class + "Accuracy");

	@XMLConfigurableConstructor({ "accuracy", "dist_type", "max_deviation", "value_offset", "communicated_accuracy" })
	public MaliciousVectoralJamEnvironmentSensorPlugin(double pAccuracy, String pDist, double pMaxDeviation,
			double pValueOffset, double pCommunicatedAccuracy) {
		_originalAccuracy = pAccuracy;
		_maxDeviation = pMaxDeviation;

		double deviation = pMaxDeviation * pAccuracy;

		int upperBound = 1;
		double lowerBound = 0.4;

		if (deviation + pAccuracy > upperBound) {
			deviation = upperBound - pAccuracy;
		}
		if (pAccuracy - deviation < lowerBound) {
			deviation = pAccuracy - lowerBound;
		}
		double accuracy = pAccuracy + _accuracyRandom.nextGaussian() * deviation;
		if (accuracy < pAccuracy - deviation) {
			accuracy = pAccuracy - deviation;
		}
		if (accuracy > pAccuracy + deviation) {
			accuracy = pAccuracy + deviation;
		}
		setAccuracy(accuracy);

		_valueOffset = pValueOffset;
		_communicatedAccuracy = pCommunicatedAccuracy;

		MeasurementDistributionTypeContainer.setDistribution(VectoralJamProperty.class,
				ProbabilityDistribution.valueOf(pDist));
	}

	@Override
	public void setHost(Host pHost) {
		_host = pHost;
	}

	@Override
	public List<EnvironmentProperty> getEnvironmentProperties() {
		if (_sis == null) {
			try {
				_sis = _host.getComponent(SiSComponent.class);
			} catch (ComponentNotAvailableException e) {
				throw new AssertionError("SiS requried!", e);
			}
		}
		
		try {
			RoadNetworkEdge edge = _sis.get().localState(SiSTypes.ROAD_EDGE, SiSRequest.NONE);
			Location location = _sis.get().localState(SiSTypes.PHY_LOCATION, SiSRequest.NONE);

			double speed = -1;
			for (RoadProperty roadProperty : edge.getActiveProperties()) {
				if (roadProperty instanceof JamProperty) {
					speed = ((JamProperty) roadProperty).getAverageSpeed();
				}
			}

			if (speed == -1) {
				speed = edge.getOriginalMaxSpeed();
			}

			double measuredSpeed = measureValue(speed);
			if (_valueOffset != 0) {
				measuredSpeed += _valueOffset;
			}

			VectoralJamProperty vectoralJamProperty = new VectoralJamProperty(location, edge);

			if (_communicatedAccuracy == 0) {
				vectoralJamProperty.setSpeed(measuredSpeed, _accuracy,
						MeasurementDistributionTypeContainer.getDistribution(VectoralJamProperty.class));
			} else {
				vectoralJamProperty.setSpeed(measuredSpeed, _communicatedAccuracy,
						MeasurementDistributionTypeContainer.getDistribution(VectoralJamProperty.class));
			}

			List<EnvironmentProperty> properties = new ArrayList<>();
			EnvironmentProperty property = vectoralJamProperty;
			properties.add(property);
			return properties;
		} catch (InformationNotAvailableException e) {
			throw new AssertionError(SiSTypes.ROAD_EDGE + " and " + SiSTypes.PHY_LOCATION + " are required!");
		}
	}

	public double measureValue(double pRealSpeed) {
		VectoralJamProperty vectoralJamProperty = new VectoralJamProperty(null, null);
		vectoralJamProperty.setSpeed(pRealSpeed, _accuracy,
				MeasurementDistributionTypeContainer.getDistribution(VectoralJamProperty.class));
		double nextDouble = _random.nextDouble();

		double currentAggregate = 0;
		for (int i = 0; i < vectoralJamProperty.getValueProbabilities().length; i++) {
			currentAggregate += vectoralJamProperty.getValueProbabilities()[i];

			if (currentAggregate > nextDouble) {
				return vectoralJamProperty.getValueAtIndex(i);
			}
		}

		throw new AssertionError("Unexpected!");
	}

	@Override
	public Class<? extends RoadProperty> getSupportedClass() {
		return JamProperty.class;
	}

	@Override
	public MaliciousVectoralJamEnvironmentSensorPlugin clone() throws CloneNotSupportedException {
		return new MaliciousVectoralJamEnvironmentSensorPlugin(_originalAccuracy,
				MeasurementDistributionTypeContainer.getDistribution(VectoralJamProperty.class).toString(),
				_maxDeviation, _valueOffset, _communicatedAccuracy);
	}

	public void setAccuracy(double pAccuracy) {
		_accuracy = pAccuracy;
	}

}
