/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


package de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.plugin;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import de.tudarmstadt.maki.simonstrator.api.Host;
import de.tudarmstadt.maki.simonstrator.api.Randoms;
import de.tudarmstadt.maki.simonstrator.api.component.ComponentNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.ProbabilityDistribution;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.EnvironmentProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.JamProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.RoadProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.properties.vectoral.VectoralJamProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.environment.data.roadcondition.RoadConditionProperty;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.Location;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSComponent;
import de.tudarmstadt.maki.simonstrator.api.component.sis.SiSRequest;
import de.tudarmstadt.maki.simonstrator.api.component.sis.exception.InformationNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.sis.type.SiSTypes;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.RoadNetworkEdge;
import de.tudarmstadt.maki.simonstrator.api.util.XMLConfigurableConstructor;

/**
 * @author Tobias Meuser (tobias.meuser@kom.tu-darmstadt.de)
 * @version 1.0 at 27.02.2018
 *
 */
public class VectoralRoadConditionEnvironmentSensorPlugin implements EnvironmentSensorPlugin {
	private Host _host;

	private SiSComponent _sis;

	private double _accuracy;

	private static Random _random = Randoms.getRandom(VectoralRoadConditionEnvironmentSensorPlugin.class);

	private static Random _accuracyRandom = Randoms.getRandom(VectoralRoadConditionEnvironmentSensorPlugin.class + "Accuracy");

	private ProbabilityDistribution _probabilityDistribution = ProbabilityDistribution.SIMPLE;

	@XMLConfigurableConstructor({ "accuracy" })
	public VectoralRoadConditionEnvironmentSensorPlugin(double pAccuracy) {
		double deviation = 0.2 * pAccuracy;
		if (deviation + pAccuracy > 1) {
			deviation = 1 - pAccuracy;
		}
		if (pAccuracy - deviation < 0.6) {
			deviation = pAccuracy - 0.6;
		}
		double accuracy = pAccuracy + _accuracyRandom.nextGaussian() * deviation;
		if (accuracy < pAccuracy - deviation) {
			accuracy = pAccuracy - deviation;
		}
		if (accuracy > pAccuracy + deviation) {
			accuracy = pAccuracy + deviation;
		}
		setAccuracy(accuracy);
	}

	@Override
	public void setHost(Host pHost) {
		_host = pHost;
	}

	@Override
	public List<EnvironmentProperty> getEnvironmentProperties() {
		if (_sis == null) {
			try {
				_sis = _host.getComponent(SiSComponent.class);
			} catch (ComponentNotAvailableException e) {
				throw new AssertionError("SiS requried!", e);
			}
		}
		
		try {
			RoadNetworkEdge edge = _sis.get().localState(SiSTypes.ROAD_EDGE, SiSRequest.NONE);
			Location location = _sis.get().localState(SiSTypes.PHY_LOCATION, SiSRequest.NONE);

			double speed = -1;
			for (RoadProperty roadProperty : edge.getActiveProperties()) {
				if (roadProperty instanceof JamProperty) {
					if (((JamProperty) roadProperty).isJammed()) {
						speed = ((JamProperty) roadProperty).getAverageSpeed();
					}
				}
			}

			if (speed == -1) {
				speed = (int) (edge.getOriginalMaxSpeed() / VectoralJamProperty.SCALING) * VectoralJamProperty.SCALING;
			}

			double measuredSpeed = measureValue(speed);
			VectoralJamProperty vectoralJamProperty = new VectoralJamProperty(location, edge);

			vectoralJamProperty.setSpeed(measuredSpeed, _accuracy, _probabilityDistribution);

			List<EnvironmentProperty> properties = new ArrayList<>();
			EnvironmentProperty property = vectoralJamProperty;
			properties.add(property);
			return properties;
		} catch (InformationNotAvailableException e) {
			throw new AssertionError(SiSTypes.ROAD_EDGE + " and " + SiSTypes.PHY_LOCATION + " are required!");
		}
	}

	public double measureValue(double pRealSpeed) {
		VectoralJamProperty vectoralJamProperty = new VectoralJamProperty(null, null);
		vectoralJamProperty.setSpeed(pRealSpeed, _accuracy, _probabilityDistribution);
		double nextDouble = _random.nextDouble();

		double currentAggregate = 0;
		for (int i = 0; i < vectoralJamProperty.getValueProbabilities().length; i++) {
			currentAggregate += vectoralJamProperty.getValueProbabilities()[i];

			if (currentAggregate > nextDouble) {
				return vectoralJamProperty.getValueAtIndex(i);
			}
		}

		throw new AssertionError("Unexpected!");
	}

	@Override
	public Class<? extends RoadProperty> getSupportedClass() {
		return RoadConditionProperty.class;
	}

	@Override
	public VectoralRoadConditionEnvironmentSensorPlugin clone() throws CloneNotSupportedException {
		return new VectoralRoadConditionEnvironmentSensorPlugin(_accuracy);
	}

	public void setAccuracy(double pAccuracy) {
		_accuracy = pAccuracy;
	}

}
