/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of Simonstrator.KOM.
 * 
 * Simonstrator.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tudarmstadt.maki.simonstrator.api.web.plotting;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Scanner;

import de.tudarmstadt.maki.simonstrator.api.web.Metric;
import de.tudarmstadt.maki.simonstrator.api.web.WebConfigurationManager;
import de.tudarmstadt.maki.simonstrator.api.web.approach.Approach;
import de.tudarmstadt.maki.simonstrator.api.web.parameter.Parameter;
import de.tudarmstadt.maki.simonstrator.api.web.plotting.Plot.PlotType;

public class DefaultPlotter implements Plotter {

	private File tempFolder = new File(System.getProperty("user.home") + "/.simonstrator/temp");

	private final String[] COLORS = new String[] { "#9e0142", "#d53e4f", "#f46d43", "#fdae61", "#fee08b", "#e6f598",
			"#abdda4", "#66c2a5", "#3288bd", "#5e4fa2" };

	private Random random = new Random();

	@Override
	public byte[] plotImage(Plot pPlot, ImageFormat pFormat) {
		if (!tempFolder.exists() && !tempFolder.mkdirs()) {
			return null;
		}

		boolean success = true;

		System.out.println("Plotter called.");

		if (success) {
			StringBuffer script = new StringBuffer();

			List<Approach> approaches = new ArrayList<>();
			for (Approach approach : pPlot.getApproaches()) {
				if (approach != null) {
					approaches.add(approach);
				} else {
					success = false;
				}
			}


			Parameter[] parameters = pPlot.getParameters();

			success &= createTemplate(script, approaches, parameters, pPlot.getMetrics(), pPlot.getFixedValues(),
					pFormat);

			if (success) {
				try {
					File tempFile = new File(tempFolder, "plotting." + random.nextLong() + ".tmp");
					if (pPlot.getType().equals(PlotType.BOX)) {
						success &= setupBoxPlot(tempFile.getName(), script, approaches, parameters);
					} else if (pPlot.getType().equals(PlotType.TIME)) {
						success &= setupTimePlot(tempFile.getName(), script, approaches, parameters);
					} else {
						success = false;
					}

					FileOutputStream scriptOutput = new FileOutputStream(tempFile);
					scriptOutput.write(script.toString().getBytes());
					scriptOutput.flush();
					scriptOutput.close();
					System.out.println("Created plotting script at : " + tempFile);
					File imageFile = new File(WebConfigurationManager.getConfig().getOptions().getPlotting().getOutputPath()
									+ "/" + tempFile.getName() + "." + pFormat.getExtension());
					if (imageFile.exists()) {
						imageFile.delete();
					}

					String command = WebConfigurationManager.getConfig().getOptions().getPlotting().getPython() + " "
							+ tempFile.getAbsolutePath();
					Process p = Runtime.getRuntime().exec(command);
					InputStream input = p.getInputStream();
					while (input.read() != -1) {
					}
					if (p.waitFor() == 0 && imageFile.exists()) {
						FileInputStream stream = new FileInputStream(imageFile);
						ByteArrayOutputStream output = new ByteArrayOutputStream();
						byte[] buffer = new byte[4096];
						int length;
						while ((length = stream.read(buffer)) > 0) {
							output.write(buffer, 0, length);
						}
						stream.close();
						imageFile.delete();
						return output.toByteArray();
					}
				} catch (IOException | InterruptedException e) {
					e.printStackTrace();
				}
			}
		}

		return null;
	}

	private boolean setupBoxPlot(String pFileName, StringBuffer pScript, List<Approach> pApproaches,
			Parameter[] pParameters) {
		pScript.append("plot = BoxPlot(db, PAPER_FULL_LEGEND_OUT, '" + pFileName + "', metric)").append("\n");
		pScript.append("\n");

		for (int i = 0; i < pParameters[0].getValues().size(); i++) {
			pScript.append("plot.addGroup([");

			for (int j = 0; j < pApproaches.size(); j++) {
				if (j != 0) {
					pScript.append(", ");
				}
				pScript.append("APPROACH_").append(j);
			}

			pScript.append("], basis_environment.combine(").append(pParameters[0].getName()).append("_").append(i)
					.append("))\n");
		}
		pScript.append("\n");

		pScript.append("plot.finish()").append("\n");

		pScript.append("exit()");

		return true;
	}

	private boolean setupTimePlot(String pFileName, StringBuffer pScript, List<Approach> pApproaches,
			Parameter[] pParameters) {
		pScript.append("plot = TimePlot(db, PAPER_FULL_LEGEND_OUT, '" + pFileName + "', metric, timescale=1000000)")
				.append("\n");
		pScript.append("\n");

		for (int i = 0; i < pApproaches.size(); i++) {
			Approach approach = pApproaches.get(i);

			switch (approach.getType()) {
			case OWN:
				pScript.append("plot.addMainSystem(");
				break;
			case BASELINE:
				pScript.append("plot.addBaseline(");
				break;
			case REFERENCE:
				pScript.append("plot.addComparison(");
				break;
			default:
				return false;
			}
			pScript.append("APPROACH_").append(i).append(", basis_environment)").append("\n");
		}
		pScript.append("\n");

		pScript.append("plot.finish()").append("\n");

		pScript.append("exit()");

		return true;
	}

	private boolean createTemplate(StringBuffer script, List<Approach> approaches,
			Parameter[] parameters, Metric[] pMetrics, Map<Parameter, String> pFixed, ImageFormat format) {
		boolean success = true;
		try {
			InputStream stream = new FileInputStream("assets/web/template.py");
			Scanner sc = new Scanner(stream);
			while (sc.hasNextLine()) {
				String line = sc.nextLine();
				line = replacePlaceholders(line);
				script.append(line).append("\n");
			}
			sc.close();
			stream.close();

			if (success) {
				defineLayout(script, format);
				script.append("\n");

				defineSystems(script, approaches);
				script.append("\n");

				for (Parameter parameter : parameters) {
					defineEnvironments(script, parameter);
					script.append("\n");
				}

				System.out.println("Metrics: " + pMetrics.length);
				for (Metric metric : pMetrics) {
					defineMetrics(script, metric);
					script.append("\n");
				}

				script.append("basis_environment = Environment({}, label=\"basis\")\n");
				ArrayList<Parameter> fixedParameters = new ArrayList<>(pFixed.keySet());
				for (int i = 0; i < fixedParameters.size(); i++) {
					script.append("fixed_parameter_").append(i).append(" = Environment({'")
							.append(fixedParameters.get(i).getName()).append("':'")
							.append(pFixed.get(fixedParameters.get(i)))
							.append("'}, label=\"basis\")\n");
					script.append("basis_environment = basis_environment.combine(fixed_parameter_").append(i)
							.append(")\n");
				}
				script.append("\n");

				return true;
			}
		} catch (IOException e) {
			e.printStackTrace();
		}

		return false;
	}

	private void defineLayout(StringBuffer pScript, ImageFormat pFormat) {
		pScript.append(
				"PAPER_FULL_LEGEND_OUT = PlotSettings(scale=0.9, heightScale=0.5, formats=['" + pFormat.getExtension()
						+ "'], titleLeft=True, legendCols=2)\n");
		pScript.append(
				"PAPER_HALF_LEGEND_OUT = PlotSettings(scale=0.5, heightScale=1.0, formats=['" + pFormat.getExtension()
						+ "'], titleLeft=True, legendCols=2)\n");

	}

	private void defineMetrics(StringBuffer pScript, Metric pMetric) {
		pScript.append("metric = SingleMetricDescription('").append(pMetric.getName())
				.append("', '" + pMetric.getDescription() + "')").append("\n");
	}

	private void defineEnvironments(StringBuffer pScript, Parameter pParameter) {
		for (int i = 0; i < pParameter.getValues().size(); i++) {
			pScript.append(pParameter.getName()).append("_").append(i).append(" = Environment({'");
			pScript.append(pParameter.getName()).append("':'").append(pParameter.getValues().get(i))
					.append("'}, label='").append(pParameter.getValues().get(i)).append("')");
			pScript.append("\n");
		}
	}

	private void defineSystems(StringBuffer pScript, List<Approach> pApproaches) {
		for (int i = 0; i < pApproaches.size(); i++) {
			pScript.append("APPROACH_").append(i).append(" = System({");
			Map<String, String> variableValues = pApproaches.get(i).getVariables();
			List<String> variables = new ArrayList<>(variableValues.keySet());
			for (int j = 0; j < variables.size(); j++) {
				if (j != 0) {
					pScript.append(", ");
				}
				pScript.append("'").append(variables.get(j)).append("':'").append(variableValues.get(variables.get(j)))
						.append("'");
			}

			pScript.append("}).withConfiguration(SystemConfiguration('").append(pApproaches.get(i).getName())
					.append("', {}, color='").append(COLORS[i]).append("'))").append("\n");

		}
	}

	private String replacePlaceholders(String pLine) {
		String result = pLine;
		result = result.replace("{{$$DATABASE_HOST$$}}", "127.0.0.1");
		result = result.replace("{{$$DATABASE_PORT$$}}", "3306");
		result = result.replace("{{$$DATABASE_USER$$}}",
				WebConfigurationManager.getConfig().getOptions().getDatabase().getUsername());
		result = result.replace("{{$$DATABASE_PASSWORD$$}}",
				WebConfigurationManager.getConfig().getOptions().getDatabase().getPassword());
		result = result.replace("{{$$DATABASE_NAME$$}}",
				WebConfigurationManager.getConfig().getOptions().getDatabase().getDatabase());
		result = result.replace("{{$$SIMONSTRATOR_PLOTTING$$}}",
				WebConfigurationManager.getConfig().getOptions().getPlotting().getPath());
		return result;
	}
}
